from example import run_example
import datetime as dt
import meteomatics.api as api


def stations_ids_example(username: str, password: str, _logger):
    now = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    startdate_station_ts = now - dt.timedelta(days=2)
    enddate_station_ts = now - dt.timedelta(hours=3)
    parameters_station_ts = ['t_2m:C', 'wind_speed_10m:ms', 'precip_1h:mm']
    interval_station_ts = dt.timedelta(hours=1)
    model_station_ts = 'mix-obs'
    wmo_stations = ['066810']  # St. Gallen
    metar_stations = ['EDDF']  # Frankfurt/Main
    mch_stations = ['STG']  # MeteoSchweiz Station St. Gallen

    _logger.info("\nstation wmo + metar ids timeseries:")
    try:
        df_sd_ids = api.query_station_timeseries(startdate_station_ts, enddate_station_ts, interval_station_ts,
                                                 parameters_station_ts, username, password, model=model_station_ts,
                                                 wmo_ids=wmo_stations, metar_ids=metar_stations,
                                                 mch_ids=mch_stations, on_invalid='fill_with_invalid',
                                                 request_type="POST", temporal_interpolation='none')
        _logger.info("Dataframe head \n" + df_sd_ids.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(stations_ids_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def grid_time_series_example(username: str, password: str, _logger):
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 3
    res_lon = 3
    startdate_ts = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    enddate_ts = startdate_ts + dt.timedelta(days=1)
    interval_ts = dt.timedelta(hours=1)
    parameters_ts = ['t_2m:C', 'precip_1h:mm']

    _logger.info("\ngrid timeseries:")
    try:
        df_grid_timeseries = api.query_grid_timeseries(startdate_ts, enddate_ts, interval_ts, parameters_ts, lat_N,
                                                       lon_W, lat_S, lon_E, res_lat, res_lon, username, password)
        _logger.info("Dataframe head \n" + df_grid_timeseries.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(grid_time_series_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def time_series_example(username: str, password: str, _logger):
    startdate_ts = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    enddate_ts = startdate_ts + dt.timedelta(days=1)
    interval_ts = dt.timedelta(hours=1)
    coordinates_ts = [(47.249297, 9.342854), (50., 10.)]
    parameters_ts = ['t_2m:C', 'precip_1h:mm']
    model = 'mix'
    ens_select = None  # e.g. 'median'
    cluster_select = None  # e.g. "cluster:1", see http://api.meteomatics.com/API-Request.html#cluster-selection

    _logger.info("\ntime series:")
    try:
        df_ts = api.query_time_series(coordinates_ts, startdate_ts, enddate_ts, interval_ts, parameters_ts,
                                      username, password, model, ens_select,
                                      cluster_select=cluster_select)
        _logger.info("Dataframe head \n" + df_ts.head().to_string())
        _logger.info("Dataframe head complete indexes \n" + df_ts.reset_index().head().to_string())
    except Exception as e:
        _logger.info("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(time_series_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def lightning_strokes_example(username: str, password: str, _logger):
    startdate_l = dt.datetime.utcnow() - dt.timedelta(days=1)
    enddate_l = dt.datetime.utcnow() - dt.timedelta(minutes=5)
    lat_N_l = 90
    lon_W_l = -180
    lat_S_l = -90
    lon_E_l = 180

    _logger.info("\nlighning strokes as csv:")
    try:
        df_lightning = api.query_lightnings(startdate_l, enddate_l, lat_N_l, lon_W_l, lat_S_l, lon_E_l, username,
                                            password)
        _logger.info("Dataframe head \n" + df_lightning.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(lightning_strokes_example)



from example import run_example
import meteomatics.api as api


def get_available_time_ranges_example(username: str, password: str, _logger):
    _logger.info("\nget available time ranges:")
    try:
        df_time_ranges = api.query_available_time_ranges(['t_2m:C', 'precip_6h:mm'], username, password,
                                                         'dwd-icon-eu')
        _logger.info("Dataframe head \n" + df_time_ranges.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(get_available_time_ranges_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def unpivoted_grid_example(username: str, password: str, _logger):
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 1
    res_lon = 1
    parameters_grid_unpiv = ['t_2m:C', 'precip_1h:mm']
    valid_dates_unpiv = [dt.datetime.utcnow(), dt.datetime.utcnow() + dt.timedelta(days=1)]

    _logger.info("\nunpivoted grid:")
    try:
        df_grid_unpivoted = api.query_grid_unpivoted(valid_dates_unpiv, parameters_grid_unpiv, lat_N, lon_W, lat_S,
                                                     lon_E, res_lat, res_lon, username, password)
        _logger.info("Dataframe head \n" + df_grid_unpivoted.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(unpivoted_grid_example)



from example import run_example
import meteomatics.api as api
from meteomatics.exceptions import Forbidden
import datetime as dt


def is_forecast_possible(username: str, password: str):
    startdate_ts = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    enddate_ts = startdate_ts + dt.timedelta(days=1)
    interval_ts = dt.timedelta(hours=8)
    coordinates_ts = [(47.249297, 9.342854)]
    parameters_ts = ['t_2m:C']

    try:
        api.query_time_series(coordinates_ts, startdate_ts, enddate_ts, interval_ts, parameters_ts, username, password)
    except Forbidden:
        return False
    return True


def is_historic_possible(username: str, password: str):
    startdate_ts = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0) - dt.timedelta(days=10)
    enddate_ts = startdate_ts + dt.timedelta(days=1)
    interval_ts = dt.timedelta(hours=8)
    coordinates_ts = [(47.249297, 9.342854)]
    parameters_ts = ['t_2m:C']

    try:
        api.query_time_series(coordinates_ts, startdate_ts, enddate_ts, interval_ts, parameters_ts, username, password)
    except Forbidden:
        return False
    return True


def is_grid_query_possible(username: str, password: str):
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 3
    res_lon = 3
    startdate_grid = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    parameter_grid = 'evapotranspiration_1h:mm'  # 't_2m:C'

    try:
        api.query_grid(startdate_grid, parameter_grid, lat_N, lon_W, lat_S, lon_E, res_lat, res_lon,
                       username, password)
    except Forbidden:
        return False
    return True


def is_model_selection_possible(username: str, password: str):
    try:
        api.query_available_time_ranges(['t_2m:C', 'precip_6h:mm'], username, password, 'dwd-icon-eu')
    except Forbidden:
        return False
    return True


def account_options_example(username: str, password: str, _logger):
    _logger.info("time series in the future: {}".format(is_forecast_possible(username, password)))
    _logger.info("grid queries: {}".format(is_grid_query_possible(username, password)))
    _logger.info("historic queries: {}".format(is_historic_possible(username, password)))
    _logger.info("model selection: {}".format(is_model_selection_possible(username, password)))


if __name__ == "__main__":
    run_example(account_options_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def get_init_dates_example(username: str, password: str, _logger):
    now = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

    _logger.info("\nget init dates:")
    try:
        df_init_dates = api.query_init_date(now, now + dt.timedelta(days=2), dt.timedelta(hours=3), 't_2m:C',
                                            username,
                                            password, 'ecmwf-ens')
        _logger.info("Dataframe head \n" + df_init_dates.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(get_init_dates_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def stations_coordinates_example(username: str, password: str, _logger):
    now = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    startdate_station_ts = now - dt.timedelta(days=2)
    enddate_station_ts = now - dt.timedelta(hours=3)
    parameters_station_ts = ['t_2m:C', 'wind_speed_10m:ms', 'precip_1h:mm']
    interval_station_ts = dt.timedelta(hours=1)
    model_station_ts = 'mix-obs'
    coordinates_station_ts = [(47.43, 9.4), (50.03, 8.52)]  # St. Gallen / Frankfurt/Main

    _logger.info("\nstation coordinates timeseries:")
    try:
        df_sd_coord = api.query_station_timeseries(startdate_station_ts, enddate_station_ts, interval_station_ts,
                                                   parameters_station_ts, username, password,
                                                   model=model_station_ts,
                                                   latlon_tuple_list=coordinates_station_ts,
                                                   on_invalid='fill_with_invalid', request_type="POST",
                                                   temporal_interpolation='none')
        _logger.info("Dataframe head \n" + df_sd_coord.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(stations_coordinates_example)



from example import run_example
import datetime as dt
import meteomatics.api as api


def grid_png_example(username: str, password: str, _logger):
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 0.1
    res_lon = 0.1
    filename_png = "grid_target.png"
    startdate_png = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    parameter_png = 't_2m:C'

    _logger.info("\ngrid as a png:")
    try:
        api.query_grid_png(filename_png, startdate_png, parameter_png, lat_N, lon_W, lat_S, lon_E, res_lat, res_lon,
                           username, password)
        _logger.info("filename = {}".format(filename_png))
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(grid_png_example)



"""
In this script a user can specify several regions for which weather data will then be plotted. It was conceived as a
regular energy market update with temperature anomaly (indicative of heating demand) and 90m wind speed (indicative of
onshore wind generation) as the time-series variables. Two sets of ouput plots are produced:
- time-series data for each region and each variable
- an animation of temperture and pressure in a given region and map projection (frames also produced)
The variables used in the time-series may be straightforwardly changed; the variables shown in the animation are
not configurable (but may be changed in the function) since, in general, a different choice of variable will imply
a different optimal plot type.

To run the script, an API account is required. Credentials should be updated in the script. A free trial can be
initiated here: https://www.meteomatics.com/en/api-trial-account-form/

I've experimented with a few different configurations but the code has not been robustly tested: let me know if you
try something that doesn't work and I'll think about an update.
"""

import shapely.geometry as sgeometry
import matplotlib.pyplot as plt
import meteomatics.api as api
import cartopy.crs as ccrs
import geopandas as gpd
from PIL import Image
import datetime as dt
import pandas as pd
import xarray as xr
import numpy as np
import functools
import glob
import time
import os

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CONFIGURATION ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# 1. What countries/regions do you want in your time-series?
# These are provided in two arguments. 'STANDARD_COUNTRIES' is a list of strings which are the names of polygons in
# geopandas' "naturalearth_lowres" built-in dataset. 'MANUAL_GEOMETRIES' is a dictionary whose keys are the names you
# want to give to your user-created polygons, and whose keys are the paths to geojson files containing those polygons.
# I advise that, if you make changes to this, you check that they have been interpreted correctly by switching
# 'SHOW_POLYGONS' to True. This will show them in your chosen map projection. Speaking of which...
STANDARD_COUNTRIES = ['United Kingdom', 'Germany', 'Netherlands', 'Italy']
MANUAL_GEOMETRIES = {'Northern France': 'polygons/northern_france.geojson'}
SHOW_POLYGONS = False

# 2. What map projection do you want to show your animated data in?
# Change entries in the dictionaries to change the bounding box for your animations. This will make a rectangular plot
# regardless of the projection. The dictionary values are then used to define the map projection. You can choose from
# any cartopy projection, but be aware that a) some projections take different arguments to define the transformations
# and b) some projections will not fill the rectangle defined by the dictionaries in certain parts of the globe (for
# instance, if you use AzimuthalEquidistant in Europe you'll see the fetched data get squeezed at the top of the plot).
TOP_LEFT = {
    'lat': 66.211199,
    'lon': -32.364446
}
BOTTOM_RIGHT = {
    'lat': 35.526388,
    'lon': 32
}
# CRS = ccrs.AzimuthalEquidistant(central_longitude=(TOP_LEFT['lon']+BOTTOM_RIGHT['lon'])/2,
#                                 central_latitude=(TOP_LEFT['lat']+BOTTOM_RIGHT['lat'])/2)
CRS = ccrs.Mercator(central_longitude=(TOP_LEFT['lon']+BOTTOM_RIGHT['lon'])/2,
                    min_latitude=BOTTOM_RIGHT['lat'],
                    max_latitude=TOP_LEFT['lat'])

# 3. What variables are you interested in?
# Currently this choice only affects time-series plots: the variables used in the animation are static (temperature and
# pressure, see animate <== make_nc). Keys should be the names you want on your axis labels, and should feature exactly
# one set of parentheses enclosing the units; values are the names of the corresponding meteomatics strings.
TIMESERIES_VARS = {
    'Temperature (C)': 't_2m:C',
    '90m Wind Speed (m/s)': 'wind_speed_90m:ms',
    'Global Radiation (W/m2)': 'global_rad:W'
}

# 4. How do you want to define the baseline (climatology) for your time-series?
# 15 years is apparently standard for energy industry; 30 years is standard in academic texts. Climatologies are all
# interpolated to 1hrly resolution, but I recommend sub-6hrly for data acquisition, since 0600 and 1800 vary between
# being daytime- and nighttime values throughout the year. Also note that the API assumes all times are UTC, so a
# smaller time-step is better for translating to other time-zones). Don't let your climatology get too out of date!
CLIMATOLOGY_START_YEAR = 2005
CLIMATOLOGY_STOP_YEAR = 2020
CLIMATOLOGY_STEP = dt.timedelta(hours=3)

# IMPORTANT! Climatologies are obtained once per new run environment and then imported from .csv in order to reduce
# runtime. The script is not clever enough to know whether you have changed TIMESERIES_VARS or the period to be covered
# by climatology, so make sure you force the recalculation of climatology whenever changes are made.
REBUILD_CLIMATOLOGY = False

# 5. Plot formatting.
#   What period do you want your output (time-series and animations) to cover?
#   What temporal resolution do you want for them both (set separately)?
#   What spatial resolution do you want in your
# Note that the animation may fail if you set the time-step very high
LEAD_TIME_DAYS = 28
START_TIME = dt.datetime.combine(dt.datetime.now().date(), dt.time(12))  # starts at 12pm on the day called
T_SERIES_LEAD = dt.timedelta(days=LEAD_TIME_DAYS)
T_SERIES_STEP = dt.timedelta(hours=1)
ANIMATION_STEP = dt.timedelta(hours=6)

# API access
USERNAME = 'energy_market_plots'
PASSWORD = 'GkQ9qqq868Lm'

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ END CONFIGURATION ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        t_start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - t_start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer


def get_geometries(standard_countries, manual_geometries, show=False):
    """
    Uses geopandas' built-in countries for those polygons which correspond to a whole country; otherwise can use custom
    geometries provided in GeoJSON format (with corresponding user-defined polygon names: see parameter docstrings).
    :param standard_countries: list of country names as they appear in naturalearth_lowres
    :param manual_geometries: dictionary of:
                polygon names : GeoJSON paths
    :param show: Bool, if True will illustrate location and shape of polygons on given map projection
    :return: dictionary of:
                polygon names (standard_countries + manual_geometries.keys) : shapely.[Multi]Polygon
    """
    countries = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
    geometries = {}
    geometries_projected = {}  # will be filled with geometries in chosen projection if show is True
    for country in standard_countries:
        geometries[country] = countries[countries.name == country].geometry.values[0]
        if show:
            geometries_projected[country] = countries[countries.name == country].to_crs(CRS.proj4_init).geometry
    for country in manual_geometries:
        geometries[country] = gpd.read_file(manual_geometries[country]).geometry.values[0]
        if show:
            geometries_projected[country] = gpd.read_file(manual_geometries[country]).to_crs(CRS.proj4_init).geometry
    if show:
        show_geometries(geometries_projected)
    return geometries


def show_geometries(geometries):  # TODO could add an option to show overlap
    """
    Plots the projected polygons
    :param geometries: shapely.[Multi]Polygon objects corresponding to polygons projected into CRS (defined in header)
    """
    fig = plt.figure()
    ax = fig.add_subplot(projection=CRS)
    ax.stock_img()
    ax.set_extent([TOP_LEFT['lon'], BOTTOM_RIGHT['lon'], TOP_LEFT['lat'], BOTTOM_RIGHT['lat']])
    for geom in geometries:
        geometries[geom].plot(ax=ax)
    plt.show()


@timer
def get_polygon_climatology(geometries, location):  # TODO could save climatology locally and only fetch if required
    if os.path.exists(os.path.join('climatologies', '{}.csv'.format(location))) and not REBUILD_CLIMATOLOGY:
        print('Climatology for {} exists: reading climatology'.format(location))
    else:
        print('New climatology required for {}.'.format(location))
        os.makedirs('climatologies', exist_ok=True)
        write_climatology(geometries, location)
    df = pd.read_csv(os.path.join('climatologies', '{}.csv'.format(location)), index_col=[0, 1])
    return df


def write_climatology(geometries, location):
    """
    This only looks hefty because of the if/else block, plus the fact that I've extended API calls over multiple lines.
    Makes a climatology dataframe by joining years together sequentially.
    :param geometries: dictionary of:
                locations : shapely.[Multi]Polygons
    :param location: the name of the location. passed because desired for name of written .csv
    """
    geometry = geometries[location]
    tuple_lists = make_tuple_lists(geometry)
    print('Fetching climatology for {}.'.format(location))
    # had to break this down because of 300s timeout. 3hrly still seems reasonable and a bit faster; could go 1hrly now
    for year in range(CLIMATOLOGY_START_YEAR, CLIMATOLOGY_STOP_YEAR):
        if year == CLIMATOLOGY_START_YEAR:
            df = api.query_polygon(
                latlon_tuple_lists=tuple_lists,
                startdate=dt.datetime(year, 1, 1),
                enddate=dt.datetime(year, 12, 31, int(24 - CLIMATOLOGY_STEP.seconds/3600.)),
                interval=CLIMATOLOGY_STEP,
                parameters=TIMESERIES_VARS.values(),
                aggregation=['mean'],
                username=USERNAME,
                password=PASSWORD,
                operator='U',  # in case of MultiPolygons
                model='ecmwf-era5',
                polygon_sampling='adaptive_grid'
            )
        else:
            df = df.append(
                api.query_polygon(
                    latlon_tuple_lists=tuple_lists,
                    startdate=dt.datetime(year, 1, 1),
                    enddate=dt.datetime(year, 12, 31, int(24 - CLIMATOLOGY_STEP.seconds/3600.)),
                    interval=CLIMATOLOGY_STEP,
                    parameters=TIMESERIES_VARS.values(),
                    aggregation=['mean'],
                    username=USERNAME,
                    password=PASSWORD,
                    operator='U',  # in case of MultiPolygons
                    model='ecmwf-era5',
                    polygon_sampling='adaptive_grid'
                )
            )
    # Query returns MultiIndexed DataFrame, but outer index is polygon1 even if geometry is a MultiPolygon (because
    # of the union operation). To make this easier to work with, I cross-section to get rid of the outer index level.
    # To future-proof, make sure that there is only one outer index - if so, remove it.
    assert len(df.index.get_level_values(0).unique()) == 1
    df = df.xs(key='polygon1')
    df = df.resample('H').interpolate(method='cubic')
    climatology = df.groupby([df.index.dayofyear, df.index.time]).mean()
    climatology.index.names = ['doy', 'time']
    climatology.to_csv(os.path.join('climatologies', '{}.csv'.format(location)))


@timer
def get_polygon_data(geometry):  # TODO could make parameters variable
    """
    Gets mean weather data for 2m temperature and 90m wind-speed (currently static) within polygon defined by geometry
    :param geometry: shapely.[Multi]Polygon
    :return: pandas.DataFrame of spatial mean time-series data for the polygon; time-series parameters defined in header
    """
    tuple_lists = make_tuple_lists(geometry)
    print('Fetching time-series data for the next {} days.'.format(LEAD_TIME_DAYS))
    df = api.query_polygon(
        latlon_tuple_lists=tuple_lists,
        startdate=START_TIME,
        enddate=START_TIME + T_SERIES_LEAD,
        interval=T_SERIES_STEP,
        parameters=TIMESERIES_VARS.values(),
        aggregation=['mean'],
        username=USERNAME,
        password=PASSWORD,
        operator='U',  # not necessary for the single polygon, but doesn't break and is necessary for multi-polygons
        model='ecmwf-vareps',
        polygon_sampling='adaptive_grid'
    )
    # Query returns MultiIndexed DataFrame, but outer index is polygon1 even if geometry is a MultiPolygon (because
    # of the union operation). To make this easier to work with, I cross-section to get rid of the outer index level.
    # To future-proof, make sure that there is only one outer index - if so, remove it.
    assert len(df.index.get_level_values(0).unique()) == 1
    df = df.xs(key='polygon1')
    return df


def make_tuple_lists(geometry):
    """
    Prepare the tuple list argument for API query from polygon data
    :param geometry: shapely.[Multi]Polygon
    :return: tuple lists (tuple list for each Polygon)
    """
    retval = []
    if type(geometry) is sgeometry.multipolygon.MultiPolygon:
        for polygon in geometry:  # this is the correct way to access Polygons within a MultiPolygon...
            retval.append(make_tuple_list(polygon))
    elif type(geometry) is sgeometry.polygon.Polygon:
        retval.append(make_tuple_list(geometry))  # ...but you can't loop over a single Polygon (which is dumb)
    else:
        # code should be written such that only [Multi]Polygons make it here; if imported modules change, will raise
        raise TypeError('geometries should contain only polygons or multipolygons')
    return retval


def make_tuple_list(polygon):
    """
    Subroutine of make_tuple_lists. Argument will always be a Polygon, since parent function loops over MultiPolygon
    :param polygon: shapely.Polygon
    :return: tuple list
    """
    tuple_list = []
    lons, lats = polygon.exterior.coords.xy
    for i in range(len(lons)):
        tuple_list.append((lats[i], lons[i]))
    return tuple_list


def get_combined_data(geometries, key, mins, maxs):
    """
    Reads the climatology and forecast data, combines them, writes the result to file and updates the min/max dicts.
    :param geometries: dict of shapely.[Multi]Polygons
    :param key: the country/region of interest
    :param mins: dict of minimum values for variables across all regions
    :param maxs: dict of maximum values for variables across all regions
    :return: (updated) dicts of minimum/maximum values for variables across all regions
    """
    clim = get_polygon_climatology(geometries, key)  # pass the key in order to name/lookup the file
    fcst = get_polygon_data(geometries[key])
    combined = combine(fcst, clim)
    for var in fcst.columns:
        mins[var] = min(mins[var], min(fcst[var]))
        maxs[var] = max(maxs[var], max(fcst[var]))
    os.makedirs('time-series', exist_ok=True)
    combined.to_csv(os.path.join('time-series', 'time-series data {}.csv'.format(key)))
    return mins, maxs


def combine(fcst, clim):
    """
    Joins the forecast and the corresponding climatology data together into one DataFrame.
    :param fcst:
    :param clim:
    :return:
    """
    relevant_clim = pd.DataFrame(index=fcst.index, columns=fcst.columns)
    for i in range(len(fcst)):
        tstamp = fcst.index[i]
        relevant_clim.loc[tstamp] = clim.loc[(tstamp.dayofyear, tstamp.time().strftime('%H:%M:%S'))]
    return pd.merge(fcst, relevant_clim, left_index=True, right_index=True, suffixes=('_forecast', '_climatology'))


def plot_timeseries(title, mins, maxs):
    """
    Generic function for plotting time-series data and anomaly from climatology of any variable (specified in header)
    :param title: the name of the region being plotted: will be the title of the plot and the name of the file
    :param mins: dictionary of global minimum values corresponding to each variable
    :param maxs: dictionary of global maximum values corresponding to each variable
    """
    df = pd.read_csv(os.path.join('time-series', 'time-series data {}.csv'.format(title)), index_col=0, parse_dates=True)
    for var in TIMESERIES_VARS:
        try:
            assert '(' in var
        except AssertionError:
            raise ValueError('Make sure the TIMESERIES_VARS dict is formatted correctly')
        mm_string = TIMESERIES_VARS[var]
        unit_index = len(var.split('(')[0])
        vmin = mins[mm_string] - (maxs[mm_string] - mins[mm_string]) * 0.1
        vmax = maxs[mm_string] + (maxs[mm_string] - mins[mm_string]) * 0.1
        f, (ax1, ax2) = plt.subplots(2, sharex='col', figsize=(12, 7))
        climatology_years = CLIMATOLOGY_STOP_YEAR - CLIMATOLOGY_START_YEAR
        df['{}_climatology'.format(mm_string)].plot(label='{}-year mean'.format(climatology_years), ax=ax1, c='b')
        df['{}_forecast'.format(mm_string)].plot(label='forecast', ax=ax1, ylim=[vmin, vmax], c='k')
        pd.Series(np.zeros(len(df.index)), index=df.index).plot(ax=ax2, linestyle='dashed', c='b')
        (df['{}_forecast'.format(mm_string)] - df['{}_climatology'.format(mm_string)]).plot(ax=ax2, c='k')
        # plot formatting options
        f.suptitle('{}'.format(title))
        ax1.set_ylabel(var)
        ax1.legend()
        ax2.set_ylabel(var[:unit_index]+'Anomaly '+var[unit_index:])
        ax1.grid(True)
        ax2.grid(True)
        f.tight_layout()
        os.makedirs(os.path.join('time-series', '{}'.format(var.replace('/', ''))), exist_ok=True)
        plt.savefig(os.path.join('time-series', '{}'.format(var.replace('/', '')), '{}.png'.format(title)))
        plt.close(f)


@ timer
def make_nc():
    """
    A single API call to retrieve all this data is too big, so I loop over days in our forecast range and stack the
    grids together. Using query_grid_timeseries is still desirable for this, since it can be manipulated into an
    xarray.Dataset very easily.
    :return: xarray.Dataset containing 2m temperature and MSLP
    """
    for day in range(LEAD_TIME_DAYS):
        query_start = START_TIME + dt.timedelta(days=day)
        df = api.query_grid_timeseries(
            startdate=query_start,
            enddate=query_start + dt.timedelta(hours=23), # ensures we don't double-count days
            interval=ANIMATION_STEP,
            parameters=['t_2m:C', 'msl_pressure:hPa'],
            lat_N=np.ceil(TOP_LEFT['lat'] + 1),
            lon_W=np.floor(TOP_LEFT['lon'] - 1),
            lat_S=np.floor(BOTTOM_RIGHT['lat'] - 1),
            lon_E=np.ceil(BOTTOM_RIGHT['lon'] + 1),
            res_lon=0.5,
            res_lat=0.5,
            username=USERNAME,
            password=PASSWORD,
            model='ecmwf-vareps'
        )
        if day == 0:
            nc = df.to_xarray()
        else:
            tmp = df.to_xarray()
            nc = xr.concat([nc, tmp], dim='validdate')
    try:
        # TODO this currently fails because Meteomatics data is float64 and to_netcdf wants float32
        nc.to_netcdf('run_{}'.format(dt.datetime.now().strftime('%Y-%m-%d_%Hh00.nc')))
    except ValueError:
        pass
    return nc


def transform_coords(nc):
    """
    I found the suggested method of adding a transorm=CRS keyword argument to plotting functions not to work with
    e.g. Azimuthal equidistant. Perhaps this is something to do with pcolor. Anyway, here's my solution: define a
    2D latitude and longitude variable which corresponds to the data variables; transform those and return them.
    :param nc: xarray dataset containing latitude and longitude coordinates
    :return: longitude and latitudes transformed to your chosen CRS
    """
    # We first need latitude- and longitude arrays of equal size. Since we may be working with a map which has different
    # x- and y-dimensions, and also since we may not be mapping to a rectilinear coordinate system, we first have to
    # make 2D arrays of each
    meshed_lon, meshed_lat = np.meshgrid(nc.lon.values, nc.lat.values)
    # We can then transform these to our target CRS. The source CRS is PlateCarree() i.e. PlateCarree with default args
    # i.e. quadratic grid, as this is the coordinate system returned from the Meteomatics API.
    transformed_output = CRS.transform_points(ccrs.PlateCarree(), meshed_lon, meshed_lat)
    # This gives us a 3D array of x, y, z; the latter will, for our purposes, be identically 0. We can subset this as
    transformed_lons = transformed_output[:, :, 0]
    transformed_lats = transformed_output[:, :, 1]
    # Note that these are again 2D (because the same latitudes are not used for all longitudes and vice versa in a
    # non-rectangular projection) and hence cannot be used as coordinates in an xarray.Dataset, but can be added as
    # additional data variables if you like.
    # This produces lat/lon fields for each element of the weather variables which allow them to be plotted in this
    # projection. The process needn't be repeated for each time-step since we assume that the spatial domain is static.
    return transformed_lons, transformed_lats


def make_frames(nc, lons, lats, animation_path):
    """
    Make all the individual images which will comprise the final gif.
    :param nc: the netCDF containing the data
    :param lons: the 2D grid of longitudes transformed to the CRS != nc.lon.values
    :param lats: the 2D grid of latitudes transformed to the CRS != nc.lat.values
    :param animation_path: location in which to save the frames
    :return:
    """
    # one advantage of having built a Dataset of all the values we're going to animate is that I can now access
    # the global min and max of all the data for our colorbar/contour levels
    mslp_min = nc['msl_pressure:hPa'].min()
    mslp_max = nc['msl_pressure:hPa'].max()
    contour_levels = np.arange(np.floor(mslp_min), np.ceil(mslp_max), 2)
    t2m_min = nc['t_2m:C'].min()
    t2m_max = nc['t_2m:C'].max()

    os.makedirs(animation_path, exist_ok=True)
    for fle in glob.glob(os.path.join(animation_path, '*')):
        os.remove(fle)  # remove all the previous animation bits and pieces

    # TODO various bits of formatting can be done
    # this will save a bunch of static images, which can of course be giffed
    for t_step in range(len(nc.validdate)):
        fig = plt.figure()
        ax = fig.add_subplot(projection=CRS)
        ax.set_extent([TOP_LEFT['lon'], BOTTOM_RIGHT['lon'], TOP_LEFT['lat'], BOTTOM_RIGHT['lat']])
        nc_step = nc.isel(validdate=t_step)
        cbar_map = ax.pcolor(lons, lats, nc_step['t_2m:C'], vmin=t2m_min, vmax=t2m_max)
        contours = ax.contour(lons, lats, nc_step['msl_pressure:hPa'], levels=contour_levels, colors='black')
        ax.clabel(contours, contours.levels[::5])
        ax.coastlines()
        plt.colorbar(cbar_map)
        plt.title(pd.to_datetime(nc.validdate[t_step].values).strftime('%Y-%m-%d %Hh00'))
        plt.savefig(os.path.join(animation_path, '{}_{}'.format(
            str(t_step).zfill(4),  # include this so that order is preserved even if we cross into January
            pd.to_datetime(nc.validdate[t_step].values).strftime('%Y-%m-%d %Hh00.png')
        )))
        plt.close(fig)


def animate(animation_path='animation'):
    """
    Makes an animation of the temperature and pressure situation over the forecast time- and region specified.
    All arguments other than the name of the animation directory are defined in the preamble.
    :param animation_path:
    :return:
    """
    nc = make_nc()
    lons, lats = transform_coords(nc)
    make_frames(nc, lons, lats, animation_path)

    # I'd have liked this to be a FuncAnimation with a slider for time control, but that seems complicated when
    # including multiple functions on a single frame (as I do with pcolor and contour) so I cheat by making a GIF
    # out of multiple images per https://bit.ly/3kRZOmB, and please check out my question https://bit.ly/3nwhNAT
    # if you have suggestions on how to improve the readability of the GIF"
    img, *imgs = [Image.open(f) for f in sorted(glob.glob(os.path.join(animation_path, '*.png')))]
    # 'duration' (below) means duration of each frame in milliseconds
    img.save(fp=os.path.join(animation_path, 'animation.gif'), format='GIF', append_images=imgs, save_all=True,
             duration=200, loop=0)


if __name__ == '__main__':
    # get the geometries of the polygons for which we want time-series plots
    geometries = get_geometries(
        standard_countries=STANDARD_COUNTRIES,
        manual_geometries=MANUAL_GEOMETRIES,
        show=SHOW_POLYGONS
    )

    # for each of those polygons, write the time-series data and get a global min/max for each variable
    mins = {var: np.inf for var in TIMESERIES_VARS.values()}
    maxs = {var: -np.inf for var in TIMESERIES_VARS.values()}
    for key in geometries:
        mins, maxs = get_combined_data(geometries, key, mins, maxs)

    # plot the time-series using these global values
    for key in geometries:
        plot_timeseries(key, mins, maxs)

    # now make the GIF
    animate('animation')  # all the other variables for this are controlled in the preamble



from example import run_example
import datetime as dt
import meteomatics.api as api


def netcdf_example(username: str, password: str, _logger):
    filename_nc = "path_netcdf/netcdf_target.nc"
    startdate_nc = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    enddate_nc = startdate_nc + dt.timedelta(days=1)
    interval_nc = dt.timedelta(days=1)
    parameter_nc = 't_2m:C'
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 3
    res_lon = 3

    _logger.info("\nnetCDF file:")
    try:
        api.query_netcdf(filename_nc, startdate_nc, enddate_nc, interval_nc, parameter_nc, lat_N, lon_W, lat_S,
                         lon_E,
                         res_lat, res_lon, username, password)
        _logger.info("filename = {}".format(filename_nc))
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(netcdf_example)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# coding:=utf-8

import argparse
import logging
import sys
from credentials import username as username_default, password as password_default

from meteomatics.logger import create_log_handler
from meteomatics._constants_ import LOGGERNAME


def run_example(example_lambda):
    parser = argparse.ArgumentParser()
    parser.add_argument('--username', default=username_default)
    parser.add_argument('--password', default=password_default)
    arguments = parser.parse_args()

    username = arguments.username
    password = arguments.password

    create_log_handler()
    logging.getLogger(LOGGERNAME).setLevel(logging.INFO)
    _logger = logging.getLogger(LOGGERNAME)

    if username is None or password is None:
        _logger.info(
        "You need to provide a username and a password, either on the command line or by inserting them in the script")
        sys.exit()

    example_lambda(username, password, _logger)



# -*- coding: utf-8 -*-

import logging
import sys

from ._constants_ import LOGGERNAME

def create_log_handler(fmt='%(asctime)s| %(levelname)s |%(message)s', stream=sys.stderr):
    formatter = logging.Formatter(fmt)

    handler = logging.StreamHandler(stream=stream)
    handler.setFormatter(formatter)

    logging.getLogger(LOGGERNAME).addHandler(handler)


def set_log_level(level):
    """Use this function to change the internal log level of the Meteomatics API Python connector.

    Note
    ----
    The log level is propagated to the Python logging library and so it's just an integer. Look at the aliases defined
    in the library (https://docs.python.org/3/library/logging.html#logging-levels) for convenience.
    """
    logger = logging.getLogger(LOGGERNAME)

    if not logger.handlers:
        create_log_handler()

    logger.setLevel(level)



from example import run_example
import datetime as dt
import meteomatics.api as api


def find_stations_example(username: str, password: str, _logger):
    now = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    startdate_station_ts = now - dt.timedelta(days=2)
    enddate_station_ts = now - dt.timedelta(hours=3)
    parameters_station_ts = ['t_2m:C', 'wind_speed_10m:ms', 'precip_1h:mm']

    _logger.info("\nfind stations:")
    try:
        met = api.query_station_list(username, password, startdate=startdate_station_ts, enddate=enddate_station_ts,
                                     parameters=parameters_station_ts)
        _logger.info("Dataframe head \n" + met.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(find_stations_example)



# -*- coding: utf-8 -*-
"""Meteomatics Weather API Connector

Visit https://www.meteomatics.com/en/api/overview/ for an overview of the API.
Checkout the examples!
If necessary, you can open an issue at https://github.com/meteomatics/python-connector-api or
write an email to support@meteomatics.com if you need further assistance.
"""
# Python 2 compatibility
from __future__ import print_function

import itertools
import logging
import os
import warnings
from functools import wraps
from io import StringIO
from typing import Any, Dict, List, Optional, Tuple, Union

import datetime as dt
import isodate
import pandas as pd
import requests
from urllib3.exceptions import InsecureRequestWarning

from ._constants_ import DEFAULT_API_BASE_URL, VERSION, TIME_SERIES_TEMPLATE, GRID_TEMPLATE, POLYGON_TEMPLATE, \
    GRID_TIME_SERIES_TEMPLATE, GRID_PNG_TEMPLATE, LIGHTNING_TEMPLATE, NETCDF_TEMPLATE, STATIONS_LIST_TEMPLATE, \
    INIT_DATE_TEMPLATE, AVAILABLE_TIME_RANGES_TEMPLATE, NA_VALUES, LOGGERNAME
from .binary_parser import BinaryParser
from .binary_reader import BinaryReader
from .exceptions import API_EXCEPTIONS, WeatherApiException
from .parsing_util import all_entries_postal, build_coordinates_str_for_polygon, build_coordinates_str, \
    build_coordinates_str_from_postal_codes, \
    build_response_params, convert_grid_binary_response_to_df, convert_lightning_response_to_df, \
    convert_polygon_response_to_df, \
    parse_date_num, parse_ens, parse_query_station_params, \
    parse_query_station_timeseries_params, \
    parse_time_series_params, parse_url_for_post_data, localize_datenum, sanitize_datetime, set_index_for_ts, \
    extract_user_limits

_logger = logging.getLogger(LOGGERNAME)


class Config:
    """Connector configuration.

    Attributes:
        VERIFY_SSL (bool): If False, SSL verification is disabled for requests.
            This can be useful for corporate environments where "secure" proxies are deployed.
        PROXIES (dict): Optional dictionary mapping protocol to the proxy URL, e.g. {'http': 'http://proxy:8080'}.
            If empty, no proxies are used.
    """
    _config = {
        "VERIFY_SSL": True,
        "PROXIES": {}
    }

    @staticmethod
    def get(item):
        return Config._config[item]

    @staticmethod
    def set(key, value):
        if key not in Config._config.keys():
            raise KeyError("Key '{}' does not exist.".format(key))
        Config._config[key] = value


def handle_ssl(func):
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not Config.get("VERIFY_SSL"):
            # Disable InsecureRequestWarnings if VERIFY_SSL is disabled.
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', InsecureRequestWarning)
                return func(*args, verify=False, **kwargs)
        return func(*args, **kwargs)

    return wrapper


def handle_proxy(func):
    """Passing the proxies dictionary to requests proxies optional argument."""
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not len(Config.get("PROXIES")) == 0:
            return func(*args, proxies=Config.get("PROXIES"), **kwargs)
        return func(*args, **kwargs)

    return wrapper


@handle_ssl
@handle_proxy
def get_request(*args: Any, **kwargs: Any) -> requests.Response:
    return requests.get(*args, **kwargs)


@handle_ssl
@handle_proxy
def post_request(*args: Any, **kwargs: Any) -> requests.Response:
    return requests.post(*args, **kwargs)


def create_path(_file: str) -> None:
    _path = os.path.dirname(_file)
    if not os.path.exists(_path) and len(_path) > 0:
        _logger.info("Create Path: {}".format(_path))
        os.makedirs(_path)


def query_api(
    url: str,
    username: str,
    password: str,
    request_type: str = "GET",
    timeout_seconds: int = 330,
    headers: Optional[Dict[str, str]] = None,
) -> requests.Response:
    # avoid mutable default for headers
    if headers is None:
        headers = {'Accept': 'application/octet-stream'}

    if request_type.lower() == "get":
        _logger.debug("Calling URL: {} (username = {})".format(url, username))
        response = get_request(url, timeout=timeout_seconds, auth=(username, password), headers=headers)
    elif request_type.lower() == "post":
        url, data = parse_url_for_post_data(url)
        _logger.debug("Calling URL: {} (username = {})".format(url, username))
        headers['Content-Type'] = "text/plain"
        response = post_request(url, timeout=timeout_seconds, auth=(username, password), headers=headers, data=data)
    else:
        raise ValueError('Unknown request_type: {}.'.format(request_type))

    if response.status_code != requests.codes.ok:
        exc = API_EXCEPTIONS[response.status_code]
        raise exc(response.text)

    return response


def query_user_limits(username: str, password: str, api_base_url: str = DEFAULT_API_BASE_URL) -> Dict[str, Tuple[int, int]]:
    """Get users usage and limits

    returns {limit[name]: (current_count, limit[value]) for limit in defined_limits}
    """
    response = get_request(api_base_url + '/user_stats_json', auth=(username, password))
    if response.status_code != requests.codes.ok:
        exc = API_EXCEPTIONS[response.status_code]
        raise exc(response.text)
    return extract_user_limits(response)


def convert_time_series_binary_response_to_df(
    bin_input: bytes,
    coordinate_list: List[Union[str, Tuple[float, float]]],
    parameters: List[str],
    station: bool = False,
    na_values: Tuple[Any, ...] = NA_VALUES,
) -> pd.DataFrame:
    df = raw_df_from_bin(bin_input, coordinate_list, parameters, na_values, station)
    # parse parameters which are queried as sql dates but arrive as date_num
    df = df.apply(lambda col: parse_date_num(col).dt.tz_localize("UTC") if col.name.endswith(":sql") else col)
    df = set_index_for_ts(df, station, coordinate_list)
    return df


def raw_df_from_bin(
    bin_input: bytes,
    coordinate_list: List[Union[str, Tuple[float, float]]],
    parameters: List[str],
    na_values: Tuple[Any, ...],
    station: bool,
) -> pd.DataFrame:
    binary_parser = BinaryParser(BinaryReader(bin_input), na_values)
    df = binary_parser.parse(parameters, station, coordinate_list)
    return df


def query_station_list(
    username: str,
    password: str,
    source: Optional[str] = None,
    parameters: Optional[List[str]] = None,
    startdate: Optional[dt.datetime] = None,
    enddate: Optional[dt.datetime] = None,
    location: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    elevation: Optional[float] = None,
    id: Optional[str] = None,
) -> pd.DataFrame:
    """Query available observation stations.

    Filters can be provided for source, parameters, date range, location (as "lat,lon"),
    elevation and station id. The query can be send to API as "GET" or "POST" request type.

    See also: https://www.meteomatics.com/en/api/request/advanced-requests/api-request-weather-station-mos-data/#find_station

    Returns:
        pandas.DataFrame: Station metadata with 'lat' and 'lon' columns added.
    """
    url_params_dict = parse_query_station_params(source, parameters, startdate, enddate, location, elevation, id)

    url = STATIONS_LIST_TEMPLATE.format(
        api_base_url=api_base_url,
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params_dict.items()])
    )
    response = query_api(url, username, password, request_type=request_type)

    sl = pd.read_csv(StringIO(response.text), sep=";")
    sl['lat'] = sl['Location Lat,Lon'].apply(lambda x: float(x.split(",")[0]))
    sl['lon'] = sl['Location Lat,Lon'].apply(lambda x: float(x.split(",")[1]))
    sl.drop('Location Lat,Lon', axis=1, inplace=True)

    return sl


def query_station_timeseries(
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameters: List[str],
    username: str,
    password: str,
    model: Optional[str] = "mix-obs",
    latlon_tuple_list: Optional[List[Tuple[float, float]]] = None,
    wmo_ids: Optional[List[str]] = None,
    mch_ids: Optional[List[str]] = None,
    general_ids: Optional[List[str]] = None,
    hash_ids: Optional[List[str]] = None,
    metar_ids: Optional[List[str]] = None,
    temporal_interpolation: Optional[str] = None,
    spatial_interpolation: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    na_values: Tuple[Any, ...] = NA_VALUES,
) -> pd.DataFrame:
    """Retrieve station data as time series for given IDs or coordinates.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    Coordinates or station identifiers (WMO ID, METAR ID) may be used.
    na_values: numeric values in the response which will be converted to NaN. Defaults to [-666, -777, -888, -999].

    See also: https://www.meteomatics.com/en/api/request/advanced-requests/api-request-weather-station-mos-data/#station_obs

    Returns:
        pandas.DataFrame: Time series indexed by timestamp (DateTimeIndex).
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    coordinates = build_coordinates_str(latlon_tuple_list, wmo_ids, metar_ids, mch_ids, general_ids, hash_ids)
    url_params_dict = parse_query_station_timeseries_params(model, on_invalid, temporal_interpolation,
                                                            spatial_interpolation)
    url = TIME_SERIES_TEMPLATE.format(
        api_base_url=api_base_url,
        coordinates=coordinates,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        parameters=",".join(parameters),
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params_dict.items()])
    )

    headers = {'Accept': 'text/csv'}
    response = query_api(url, username, password, request_type=request_type, headers=headers)
    coordinates_list = coordinates.split("+")
    return convert_time_series_binary_response_to_df(response.content, coordinates_list, parameters,
                                                     station=True, na_values=na_values)


def query_special_locations_timeseries(
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameters: List[str],
    username: str,
    password: str,
    model: Optional[str] = "mix",
    postal_codes: Optional[Dict[str, List[int]]] = None,
    temporal_interpolation: Optional[str] = None,
    spatial_interpolation: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    na_values: Tuple[Any, ...] = NA_VALUES,
) -> pd.DataFrame:
    """Retrieve time series for locations specified by postal codes.

    postal_codes should be a dict mapping country codes to lists of postal codes,
    e.g. {'DE': [71679, 70173]}.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    na_values: numeric values in the response which will be converted to NaN. Defaults to [-666, -777, -888, -999].

    Returns:
        pandas.DataFrame: Time series indexed by timestamp (DateTimeIndex).
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    coordinates = build_coordinates_str_from_postal_codes(postal_codes)
    url_params = parse_query_station_timeseries_params(model, on_invalid, temporal_interpolation, spatial_interpolation)
    url = TIME_SERIES_TEMPLATE.format(
        api_base_url=api_base_url,
        coordinates=coordinates,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        parameters=",".join(parameters),
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params.items()])
    )

    headers = {'Accept': 'text/csv'}
    response = query_api(url, username, password, request_type=request_type, headers=headers)
    coordinates_list = coordinates.split("+")
    return convert_time_series_binary_response_to_df(response.content, coordinates_list, parameters, station=True,
                                                     na_values=na_values)


def query_time_series(
    coordinate_list: List[Union[str, Tuple[float, float]]],
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameters: List[str],
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    cluster_select: Optional[str] = None,
    na_values: Tuple[Any, ...] = NA_VALUES,
    **kwargs: Any
) -> pd.DataFrame:
    """Retrieve a time series for one or more coordinates.

    coordinate_list should be either one (lat,lon) tuples or postal-code strings
    (i.e. must contain 'postal_' prefix), mixing the two doesn't work

    If ensemble selection is requested, additional columns for ensemble members may be returned.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    na_values: numeric values in the response which will be converted to NaN. Defaults to [-666, -777, -888, -999].

    Returns:
        pandas.DataFrame: Time series indexed by timestamp (DateTimeIndex).
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    extended_params = parameters if ens_select is None else build_response_params(parameters, parse_ens(ens_select))
    url_params = parse_time_series_params(model, ens_select, cluster_select, interp_select, on_invalid, kwargs)

    is_postal = all_entries_postal(coordinate_list)
    coordinate_list_str = '+'.join(coordinate_list) if is_postal else "+".join(
        ["{},{}".format(*latlon_tuple) for latlon_tuple in coordinate_list])

    url = TIME_SERIES_TEMPLATE.format(
        api_base_url=api_base_url,
        coordinates=coordinate_list_str,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        parameters=",".join(parameters),
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params.items()])
    )

    response = query_api(url, username, password, request_type=request_type)
    df = convert_time_series_binary_response_to_df(response.content, coordinate_list, extended_params,
                                                   na_values=na_values)

    return df


def query_grid(
    startdate: dt.datetime,
    parameter_grid: str,
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    na_values: Tuple[Any, ...] = NA_VALUES,
    **kwargs: Any
) -> pd.DataFrame:
    """Retrieve a rectangular grid for a single valid date.

    Start datetime must be timezone-aware in UTC or will be interpreted as UTC.

    Parameters:
        lat_N, lon_W, lat_S, lon_E (float): Bounding box coordinates (north, west, south, east).
        res_lat, res_lon (float): Spatial resolution in degrees for latitude and longitude.
        ens_select (Optional[str]): Ensemble selection string (returns ensemble members if used).
        na_values (tuple): Values to be interpreted as missing (converted to NaN).

    Returns:
        pd.DataFrame with values of the parameter, lat as index, lon as columns
    """
    # interpret time as UTC
    startdate = sanitize_datetime(startdate)

    # build URL
    url_params = parse_time_series_params(model=model, ens_select=ens_select, cluster_select=None,
                                          interp_select=interp_select, on_invalid=on_invalid, kwargs=kwargs)
    url = GRID_TEMPLATE.format(
        api_base_url=api_base_url,
        startdate=startdate.isoformat(),
        parameter_grid=parameter_grid,
        lat_N=lat_N,
        lon_W=lon_W,
        lat_S=lat_S,
        lon_E=lon_E,
        res_lat=res_lat,
        res_lon=res_lon,
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params.items()])
    )

    response = query_api(url, username, password, request_type=request_type)
    return convert_grid_binary_response_to_df(response.content, parameter_grid, na_values=na_values)

def query_grid_unpivoted(
    valid_dates: List[dt.datetime],
    parameters: List[str],
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    on_invalid: Optional[str] = None,
    request_type: str = "GET",
    na_values: Tuple[Any, ...] = NA_VALUES,
) -> pd.DataFrame:
    """Retrieve a rectangular grid for a more then one valid date.

    Internally calls query_grid for each of `valid_dates`. Then parses the
    responses into a data frame with MultiIndex(valid_date, lat, lon) and a
    column per each of requested `parameters`.

    """
    idxcols = ['valid_date', 'lat', 'lon']
    vd_dfs = []

    for valid_date in valid_dates:
        vd_df = None
        for parameter in parameters:

            dmo = query_grid(valid_date, parameter, lat_N, lon_W, lat_S, lon_E, res_lat, res_lon, username, password,
                             model, ens_select, interp_select, on_invalid=on_invalid, request_type=request_type, na_values=na_values)

            df = pd.melt(dmo.reset_index(), id_vars='lat', var_name='lon', value_name=parameter)
            df['valid_date'] = valid_date
            df.lat = df.lat.apply(float)
            df.lon = df.lon.apply(float)

            if vd_df is None:
                vd_df = df
            else:
                vd_df = vd_df.merge(df, on=idxcols)

        vd_dfs.append(vd_df)

    data = pd.concat(vd_dfs)

    # sort_values might not available in older pandas versions
    try:
        data.sort_values(idxcols, inplace=True)
    except AttributeError:
        data.sort(idxcols, inplace=True)

    data.set_index(idxcols, inplace=True)

    return data


def query_grid_timeseries(
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameters: List[str],
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    na_values: Tuple[Any, ...] = NA_VALUES,
    **kwargs: Any
) -> pd.DataFrame:
    """Retrieve a grid (aka Rectangle) time series as tabular data.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    If ensemble selection is requested, additional columns for ensemble members may be returned.
    na_values: numeric values in the response which will be converted to NaN. Defaults to [-666, -777, -888, -999].

    See also: https://www.meteomatics.com/en/api/request/required-parameters/coordinate-description/

    Returns:
        pandas.DataFrame: Time-indexed grid time series; each column corresponds to a grid point.
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    url_params = parse_time_series_params(model=model, ens_select=ens_select, cluster_select=None,
                                          interp_select=interp_select, on_invalid=on_invalid, kwargs=kwargs)
    url = GRID_TIME_SERIES_TEMPLATE.format(
        api_base_url=api_base_url,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        lat_N=lat_N,
        lon_W=lon_W,
        lat_S=lat_S,
        lon_E=lon_E,
        res_lat=res_lat,
        res_lon=res_lon,
        parameters=",".join(parameters),
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params.items()])
    )

    response = query_api(url, username, password, request_type=request_type)

    lats = arange(lat_S, lat_N, res_lat)
    lons = arange(lon_W, lon_E, res_lon)

    latlon_tuple_list = list(itertools.product(lats, lons))
    df = convert_time_series_binary_response_to_df(response.content, latlon_tuple_list, parameters, na_values=na_values)

    return df


def query_polygon(
    latlon_tuple_lists: List[List[Tuple[float, float]]],
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameters: List[str],
    aggregation: List[str],
    username: str,
    password: str,
    operator: Optional[str] = None,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    on_invalid: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    cluster_select: Optional[str] = None,
    **kwargs: Any
) -> pd.DataFrame:
    """Retrieve time series aggregated over polygon geometries.

    Polygons must be lists of (lat,lon) tuples. Aggregation may be specified per polygon (mean, max, min, etc.).
    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    If ensemble selection is requested, additional columns for ensemble members may be returned.
    na_values: numeric values in the response which will be converted to NaN. Defaults to [-666, -777, -888, -999].

    See also: https://www.meteomatics.com/en/api/request/required-parameters/coordinate-description/

    Returns:
        pandas.DataFrame: Aggregated time series indexed by timestamp (DateTimeIndex).
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    url_params_dict = parse_time_series_params(model=model, ens_select=ens_select, cluster_select=cluster_select,
                                               interp_select=interp_select, on_invalid=on_invalid, kwargs=kwargs)
    coordinates = build_coordinates_str_for_polygon(latlon_tuple_lists, aggregation, operator)
    url = POLYGON_TEMPLATE.format(
        api_base_url=api_base_url,
        coordinates_aggregation=coordinates,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        parameters=",".join(parameters),
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params_dict.items()])
    )

    response = query_api(url, username, password, request_type=request_type)
    df = convert_polygon_response_to_df(response.text)
    return df


def query_lightnings(
    startdate: dt.datetime,
    enddate: dt.datetime,
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    username: str,
    password: str,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    model: str = "mix",
) -> pd.DataFrame:
    """Query lightning strokes in a bounding box and time range.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    The response is parsed into a DataFrame with one row per lightning stroke
    and relevant metadata columns.

    See also: https://www.meteomatics.com/en/api/available-parameters/weather-parameter/lightnings/#lightningdensity

    Returns:
        pandas.DataFrame: Lightning stroke records for the requested area and period.
    """

    # interpret time as UTC
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    url = LIGHTNING_TEMPLATE.format(
        api_base_url=api_base_url,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        lat_N=lat_N,
        lon_W=lon_W,
        lat_S=lat_S,
        lon_E=lon_E,
        source=model
    )

    headers = {'Accept': 'text/csv'}
    response = query_api(url, username, password, request_type=request_type, headers=headers)

    return convert_lightning_response_to_df(response.text)


def query_netcdf(
    filename: str,
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameter_netcdf: str,
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
    cluster_select: Optional[str] = None,
) -> None:
    """Download a netCDF file from the Meteomatics API and save to `filename`.

    Request returns binary netCDF content which is written to the given path.
    Note: the target directory will be created if it does not exist.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    If ensemble selection is requested, additional columns for ensemble members may be returned.

    Returns:
        None
    """

    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    # build URL
    url_params_dict = parse_time_series_params(model=model, ens_select=ens_select, cluster_select=cluster_select,
                                               interp_select=interp_select)
    url = NETCDF_TEMPLATE.format(
        api_base_url=api_base_url,
        startdate=startdate.isoformat(),
        enddate=enddate.isoformat(),
        interval=isodate.duration_isoformat(interval),
        parameter_netcdf=parameter_netcdf,
        lat_N=lat_N,
        lon_W=lon_W,
        lat_S=lat_S,
        lon_E=lon_E,
        res_lat=res_lat,
        res_lon=res_lon,
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params_dict.items()])
    )

    headers = {'Accept': 'application/netcdf'}
    response = query_api(url, username, password, request_type=request_type, headers=headers)

    # Check if target directory exists
    create_path(filename)

    # save to the specified filename
    with open(filename, 'wb') as f:
        _logger.debug('Create File {}'.format(filename))
        for chunk in response.iter_content(chunk_size=1024):
            f.write(chunk)

    return


def query_init_date(
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameter: str,
    username: str,
    password: str,
    model: str,
    api_base_url: str = DEFAULT_API_BASE_URL,
) -> pd.DataFrame:
    # set time zone info to UTC if necessary
    startdate = sanitize_datetime(startdate)
    enddate = sanitize_datetime(enddate)

    interval_string = "{}--{}:{}".format(startdate.isoformat(),
                                         enddate.isoformat(),
                                         isodate.duration_isoformat(interval))

    url = INIT_DATE_TEMPLATE.format(api_base_url=api_base_url,
                                    model=model, interval_string=interval_string, parameter=parameter)

    headers = {'Accept': 'text/csv'}
    response = query_api(url, username, password, request_type='GET', headers=headers)

    try:
        df = pd.read_csv(
            StringIO(response.text),
            sep=";",
            header=0,
            encoding="utf-8",
            index_col=0,
            na_values=["0000-00-00T00:00:00Z"],
            parse_dates=[0, 1]
        )
    except Exception:
        raise WeatherApiException(response.text)

    df = localize_datenum(df)

    return df


def query_available_time_ranges(
    parameters: List[str],
    username: str,
    password: str,
    model: str,
    api_base_url: str = DEFAULT_API_BASE_URL,
) -> pd.DataFrame:
    url = AVAILABLE_TIME_RANGES_TEMPLATE.format(api_base_url=api_base_url,
                                                model=model, parameters=",".join(parameters))

    headers = {'Accept': 'text/csv'}
    response = query_api(url, username, password, request_type='GET', headers=headers)

    try:
        df = pd.read_csv(
            StringIO(response.text),
            sep=";",
            header=0,
            encoding="utf-8",
            index_col=0,
            na_values=["0000-00-00T00:00:00Z"],
            parse_dates=['min_date', 'max_date']
        )
    except Exception:
        raise WeatherApiException(response.text)

    return df


def query_grid_png(
    filename: str,
    startdate: dt.datetime,
    parameter_grid: str,
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
) -> None:
    """Request a PNG image for a single grid timestamp and save to `filename`.

    The PNG is produced server-side from the specified grid parameter and written to the
    specified file path.

    Start datetime must be timezone-aware in UTC or will be interpreted as UTC.
    If ensemble selection is requested, additional columns for ensemble members may be returned.

    Returns:
        None
    """

    # interpret time as UTC
    startdate = sanitize_datetime(startdate)

    # build URL

    url_params = dict()
    url_params['connector'] = VERSION
    if model is not None:
        url_params['model'] = model

    if ens_select is not None:
        url_params['ens_select'] = ens_select

    if interp_select is not None:
        url_params['interp_select'] = interp_select

    url = GRID_PNG_TEMPLATE.format(
        api_base_url=api_base_url,
        startdate=startdate.isoformat(),
        parameter_grid=parameter_grid,
        lat_N=lat_N,
        lon_W=lon_W,
        lat_S=lat_S,
        lon_E=lon_E,
        res_lat=res_lat,
        res_lon=res_lon,
        urlParams="&".join(["{}={}".format(k, v) for k, v in url_params.items()])
    )

    headers = {'Accept': 'image/png'}
    response = query_api(url, username, password, request_type=request_type, headers=headers)

    # save to the specified filename
    with open(filename, 'wb') as f:
        _logger.debug('Create File {}'.format(filename))
        for chunk in response.iter_content(chunk_size=1024):
            f.write(chunk)

    return


def query_png_timeseries(
    prefixpath: str,
    startdate: dt.datetime,
    enddate: dt.datetime,
    interval: dt.timedelta,
    parameter: str,
    lat_N: float,
    lon_W: float,
    lat_S: float,
    lon_E: float,
    res_lat: float,
    res_lon: float,
    username: str,
    password: str,
    model: Optional[str] = None,
    ens_select: Optional[str] = None,
    interp_select: Optional[str] = None,
    api_base_url: str = DEFAULT_API_BASE_URL,
    request_type: str = "GET",
) -> None:
    """Download a sequence of PNG images for a time range and save them under `prefixpath`.

    Each image filename is constructed from the parameter and timestamp. The function iterates
    from startdate to enddate with the provided interval and saves each PNG using query_grid_png.

    Start and end datetimes must be timezone-aware in UTC or will be interpreted as UTC.
    If ensemble selection is requested, additional columns for ensemble members may be returned.

    Returns:
        None
    """

    # iterate over all requested dates
    this_date = startdate
    while this_date <= enddate:
        # construct filename
        if len(prefixpath) > 0:
            filename = prefixpath + '/' + parameter.replace(':', '_') + '_' + this_date.strftime(
                '%Y%m%d_%H%M%S') + '.png'
        else:
            filename = parameter.replace(':', '_') + '_' + this_date.strftime(
                '%Y%m%d_%H%M%S') + '.png'

        # query base method
        query_grid_png(filename, this_date, parameter, lat_N, lon_W, lat_S, lon_E, res_lat,
                       res_lon, username, password, model, ens_select, interp_select,
                       api_base_url, request_type=request_type)

        this_date = this_date + interval

    return


def arange(start: float, stop: float, step: float) -> List[float]:
    data: List[float] = []
    if start >= stop:
        return data
    while start <= stop:
        data.append(start)
        start = round(start + step, 10)
    return data



import datetime
import datetime as dt

import pytz

from example import run_example
from meteomatics._constants_ import DEFAULT_API_BASE_URL, NA_VALUES
from meteomatics.api import query_api, raw_df_from_bin


def data_changer(df):
    value = 10
    startdate = dt.datetime(2025, 4, 28, 1, 0, tzinfo=pytz.utc)
    enddate = dt.datetime(2025, 4, 28, 5, 0, tzinfo=pytz.utc)
    df.loc[:, :] = 0
    df = df.reset_index()
    idx_start = list(df[df['validdate'] == startdate].index)
    idx_end = list(df[df['validdate'] == enddate].index)
    for i in range(len(idx_start)):
        df.loc[idx_start[i]: idx_end[i], 'wind_speed_10m:ms'] = value
    df = df.set_index(['lat', 'lon', 'validdate'], drop=True)
    return df


def dataframe_directly_from_binary_data(username: str, password: str, _logger):
    startdate_ts = dt.datetime.now(tz=pytz.utc).replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
    enddate_ts = (startdate_ts + dt.timedelta(days=1))
    url = f'{DEFAULT_API_BASE_URL}/{startdate_ts.isoformat()}Z--{enddate_ts.isoformat()}Z:PT1H/t_2m:C,wind_speed_10m:ms/47.249297,9.342854+50.0,10.0/bin?connector=python_v2.11.4&model=mix'

    # The following data are needed to decode the api response
    coordinates_ts = [(47.249297, 9.342854), (50., 10.)]
    parameters_ts = ['t_2m:C', 'wind_speed_10m:ms']

    try:
        response = query_api(url, username, password, request_type='GET')
        df = raw_df_from_bin(response.content, coordinates_ts, parameters_ts, na_values=NA_VALUES, station=False)

        _logger.info("\nData frame shape: {}".format(df.shape))
        _logger.info("Dataframe head \n" + df.head().to_string())
        _logger.info("Dataframe head complete indexes \n" + df.reset_index().head().to_string())
    except Exception as e:
        _logger.info("Failed, the exception is {}".format(e))
        _logger.info("URL: {}".format(url))
        return False
    return True


if __name__ == "__main__":
    run_example(dataframe_directly_from_binary_data)



from example import run_example
import datetime as dt
import meteomatics.api as api


def grid_example(username: str, password: str, _logger):
    lat_N = 50
    lon_W = -15
    lat_S = 20
    lon_E = 10
    res_lat = 1
    res_lon = 1
    startdate_grid = dt.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    parameter_grid = 'evapotranspiration_1h:mm'  # 't_2m:C'

    _logger.info("\ngrid:")
    try:
        df_grid = api.query_grid(startdate_grid, parameter_grid, lat_N, lon_W, lat_S, lon_E, res_lat, res_lon,
                                 username, password)
        _logger.info("Dataframe head \n" + df_grid.head().to_string())
    except Exception as e:
        _logger.error("Failed, the exception is {}".format(e))
        return False
    return True


if __name__ == "__main__":
    run_example(grid_example)
